print("import: 'cvxpy'")
import cvxpy

print("import: 'cvxpy.atoms'")
import cvxpy.atoms

print("import: 'cvxpy.atoms.affine'")
import cvxpy.atoms.affine

print("import: 'cvxpy.atoms.elementwise'")
import cvxpy.atoms.elementwise

print("import: 'cvxpy.cvxcore'")
import cvxpy.cvxcore

print("import: 'cvxpy.cvxcore.python'")
import cvxpy.cvxcore.python

print("import: 'cvxpy.constraints'")
import cvxpy.constraints

print("import: 'cvxpy.expressions'")
import cvxpy.expressions

print("import: 'cvxpy.expressions.constants'")
import cvxpy.expressions.constants

print("import: 'cvxpy.interface'")
import cvxpy.interface

print("import: 'cvxpy.interface.numpy_interface'")
import cvxpy.interface.numpy_interface

print("import: 'cvxpy.lin_ops'")
import cvxpy.lin_ops

print("import: 'cvxpy.problems'")
import cvxpy.problems

print("import: 'cvxpy.reductions'")
import cvxpy.reductions

print("import: 'cvxpy.reductions.complex2real'")
import cvxpy.reductions.complex2real

print("import: 'cvxpy.reductions.complex2real.atom_canonicalizers'")
import cvxpy.reductions.complex2real.atom_canonicalizers

print("import: 'cvxpy.reductions.dcp2cone'")
import cvxpy.reductions.dcp2cone

print("import: 'cvxpy.reductions.dcp2cone.atom_canonicalizers'")
import cvxpy.reductions.dcp2cone.atom_canonicalizers

print("import: 'cvxpy.reductions.eliminate_pwl'")
import cvxpy.reductions.eliminate_pwl

print("import: 'cvxpy.reductions.eliminate_pwl.atom_canonicalizers'")
import cvxpy.reductions.eliminate_pwl.atom_canonicalizers

print("import: 'cvxpy.reductions.qp2quad_form'")
import cvxpy.reductions.qp2quad_form

print("import: 'cvxpy.reductions.qp2quad_form.atom_canonicalizers'")
import cvxpy.reductions.qp2quad_form.atom_canonicalizers

print("import: 'cvxpy.reductions.eliminate_pwl.atom_canonicalizers'")
import cvxpy.reductions.eliminate_pwl.atom_canonicalizers

print("import: 'cvxpy.reductions.solvers'")
import cvxpy.reductions.solvers

print("import: 'cvxpy.reductions.solvers.conic_solvers'")
import cvxpy.reductions.solvers.conic_solvers

print("import: 'cvxpy.reductions.solvers.qp_solvers'")
import cvxpy.reductions.solvers.qp_solvers

print("import: 'cvxpy.reductions.solvers.lp_solvers'")
import cvxpy.reductions.solvers.lp_solvers

print("import: 'cvxpy.tests'")
import cvxpy.tests

print("import: 'cvxpy.transforms'")
import cvxpy.transforms

print("import: 'cvxpy.utilities'")
import cvxpy.utilities

print("import: 'cvxpy.cvxcore.python'")
import cvxpy.cvxcore.python

